use crate::errors::Error;
use crate::models::characters::{Character, CharacterDataType, NewCharacter, StrippedCharacter};
use crate::models::users::{NewUser, User};
use sqlx::database::HasValueRef;
use sqlx::sqlite::{SqlitePool, SqlitePoolOptions};
use sqlx::Decode;

pub type TenebrousDbConn<'a> = rocket::State<'a, SqlitePool>;

pub(crate) async fn create_pool() -> Result<SqlitePool, crate::errors::Error> {
    SqlitePoolOptions::new()
        .max_connections(5)
        .connect("tenebrous.sqlite")
        .await
        .map_err(|e| e.into())
}

#[rocket::async_trait]
pub(crate) trait Dao {
    async fn load_user_by_id(&self, id: i32) -> Result<Option<User>, Error>;

    async fn load_user(&self, for_username: &str) -> Result<Option<User>, Error>;

    async fn insert_user(&self, new_user: NewUser<'_>) -> Result<User, Error>;

    async fn load_character_list(&self, for_user_id: i32) -> Result<Vec<StrippedCharacter>, Error>;

    async fn load_character(&self, character_id: i32) -> Result<Option<Character>, Error>;

    async fn insert_character(&self, new_character: NewCharacter<'_>) -> Result<(), Error>;

    async fn update_character_sheet<'a>(&self, character: &'a Character) -> Result<(), Error>;
}

//TODO is:
// - reimplement all methods
// - remove owned data
// - return sqlx result so we don't have .into() calls everywhere
// - use compile time queries
// - find replacement for diesel migrations
#[rocket::async_trait]
impl Dao for SqlitePool {
    async fn load_user_by_id(&self, user_id: i32) -> Result<Option<User>, Error> {
        sqlx::query_as::<_, User>("SELECT * FROM users WHERE id = ?")
            .bind(user_id)
            .fetch_optional(self)
            .await
            .map_err(|e| e.into())
    }

    async fn load_user(&self, for_username: &str) -> Result<Option<User>, Error> {
        sqlx::query_as::<_, User>("SELECT * FROM users WHERE username = ?")
            .bind(for_username)
            .fetch_optional(self)
            .await
            .map_err(|e| e.into())
    }

    async fn insert_user(&self, new_user: NewUser<'_>) -> Result<User, Error> {
        sqlx::query("INSERT INTO users (username, password) values (?, ?)")
            .bind(new_user.username)
            .bind(new_user.password)
            .execute(self)
            .await?;

        self.load_user(new_user.username)
            .await
            .and_then(|user| user.ok_or(Error::NotFound))
    }

    async fn load_character_list(&self, for_user_id: i32) -> Result<Vec<StrippedCharacter>, Error> {
        sqlx::query_as::<_, StrippedCharacter>(
            "SELECT id, user_id, viewable, character_name, data_type, data_version
             FROM characters WHERE user_id = ?",
        )
        .bind(for_user_id)
        .fetch_all(self)
        .await
        .map_err(|e| e.into())
    }

    async fn load_character(&self, character_id: i32) -> Result<Option<Character>, Error> {
        sqlx::query_as::<_, Character>(
            "SELECT id, user_id, viewable, character_name, data_type, data_version, data
             FROM characters WHERE id = ?",
        )
        .bind(character_id)
        .fetch_optional(self)
        .await
        .map_err(|e| e.into())
    }

    async fn insert_character(&self, new_character: NewCharacter<'_>) -> Result<(), Error> {
        sqlx::query(
            "INSERT INTO characters
                    (user_id, viewable, character_name, data_type, data_version, data)
                    values (?, ?, ?, ?, ?, ?)",
        )
        .bind(new_character.user_id)
        .bind(new_character.viewable)
        .bind(new_character.character_name)
        .bind(new_character.data_type)
        .bind(new_character.data_version)
        .bind(new_character.data)
        .execute(self)
        .await?;

        Ok(())
    }

    async fn update_character_sheet<'a>(&self, character: &'a Character) -> Result<(), Error> {
        sqlx::query("UPDATE characters set data = ? where id = ?")
            .bind(&character.data)
            .bind(character.id)
            .execute(self)
            .await?;

        Ok(())
    }
}
